		title	'Turbo Pascal 5.0 - CRT Unit'
		page	60,120
		radix	16

UniCRT		=	1

;---------------------------------- CRT modes ---------------------------------

BW40		=	0		; 40x25 B/W on Color Adapter
CO40		=	1		; 40x25 Color on Color Adapter
BW80		=	2		; 80x25 B/W on Color Adapter
CO80		=	3		; 80x25 Color on Color Adapter
Mono		=	7		; 80x25 on Monochrome Adapter
Font8x8 	=	100		; Add-in for ROM font

;-------------------- Mode constants for 3.0 compatibility --------------------

C40		=	CO40
C80		=	CO80

;----------------- Foreground and background color constants ------------------

Black		=	0
Blue		=	1
Green		=	2
Cyan		=	3
Red		=	4
Magenta 	=	5
Brown		=	6
LightGray	=	7

;------------------------- Foreground color constants -------------------------

DarkGray	=	8
LightBlue	=	9
LightGreen	=	0a
LightCyan	=	0bh
LightRed	=	0c
LightMagenta	=	0dh
Yellow		=	0e
White		=	0f
Blink		=	80		; Add-in for blinking

;------------------------------- BIOS variables -------------------------------

BiosData	=	0040h
CurrMode	=	byte ptr ds:[0049]
ScrnWdth	=	byte ptr ds:[004a]
CursLocn	=	word ptr ds:[0050]
Addr6845	=	word ptr ds:[0063]
EgaInfo 	=	word ptr es:[0087]

;--------------------------- DOS Filemode constants ---------------------------

fmClosed	=	0d7b0
fmInput 	=	0d7b1
fmOutput	=	0d7b2
fmInOut 	=	0d7b3

;----------------------------------- Others -----------------------------------

bell		=	7
bs		=	8
cr		=	0dh
lf		=	0ah
EscCh		=	1bh

		if	UniCRT
tab		=	9
formfeed	=	0ch
clreoln 	=	18h
rubout		=	7fh
		endif

CtrlA		=	1
CtrlD		=	4
CtrlF		=	6
CtrlS		=	13h
CtrlZ		=	1ah

data		segment word public

extrn		CheckBreak : Byte	; Enable Ctrl-Break
extrn		CheckEOF : Byte 	; Enable Ctrl-Z
extrn		DirectVideo : Byte	; Enable direct video addressing
extrn		CheckSnow : Byte	; Enable snow filtering
extrn		LastMode : Word 	; Current text mode
extrn		TextAttr : Byte 	; Current text attribute
extrn		WindMin : Word		; Window upper left coordinates
extrn		WindMax : Word		; Window lower right coordinates

extrn		Input : Byte
extrn		Output : Byte

WindMinX	=	byte ptr WindMin
WindMinY	=	byte ptr WindMin+1
WindMaxX	=	byte ptr WindMax
WindMaxY	=	byte ptr WindMax+1

SaveWindMax	dw	0		; Original window sizes
SaveMode	db	0
CharBuff	db	0		; Second character for zero scan chars
CtrlBrkFlag	db	0		; Ctrl-Break pending

		if	UniCRT
ESCflag 	db	0
		endif

data		ends


code		segment byte public
		assume cs:code,ds:data

		public	InitPart2, AssignCrt, KeyPressed, ReadKey
		public	GotoXY,    WhereX,    WhereY,     ClrScr,
                public  ClrEol,    NormVideo

                if      UniCRT
                public  SetTextAttr
                public  SetCursor
                else
                public  TextMode, Window, InsLine, DelLine
                public  TextColor,TextBackground,LowVideo,HighVideo
                endif

InitPart2	proc near		; proc near in original Crt ???
		mov	ah,0fh
		call	CallBios	; Get Video Info
		cmp	al,Mono
		jz	l0323
		cmp	al,Co80
		jbe	l0323		; CGA modes
		mov	ax,Co80
		call	EgaMode
l0323:		call	SetMode
		mov	ah,8
		xor	bh,bh
		call	CallBios	; Get character / attribute at cursor
		mov	al,ah		; al = ah = Attr
		and	al,7fh		; Remove blink attr
		mov	SaveMode,al
		mov	TextAttr,al
		xor	ax,ax
		mov	CheckEOF,al	; CheckEOF := False
		mov	CharBuff,al	; CharBuff := Null
		mov	CtrlBrkFlag,al	; CtrlBrkFlag := False

                if      UniCRT
                mov     ESCflag, al     ; ESCflag:= 0;
                endif

		inc	ax
		mov	CheckBreak,al	; CheckBreak := True

		push	ds
		push	cs
		pop	ds
		mov	dx,offset Int1B
		mov	ax,251bh
		int	21h
		pop	ds
		ret
InitPart2	endp

EgaMode:	mov	dx,BiosData
		mov	es,dx
		and	byte ptr es:EgaInfo,0feh
		cmp	al,Mono
		jz	l038e
		cmp	al,4
		jb	l038e
		mov	al,Co80
l038e:		push	ax
		mov	ah,0
l0391:		call	CallBios	; Set video mode in al
		pop	ax
		or	ah,ah
		jz	l03c6
		mov	ax,1112h
		mov	bl,0
		call	CallBios	; Set mode, load double charset in 0
		mov	ax,1130h
		mov	bh,0
		mov	dl,0		; Get table ptr and info,
		call	CallBios	; cx-bytes, dl = rows per char
		cmp	dl,2ah
		jnz	l03c6
		or	byte ptr es:EgaInfo,1
		mov	ax,100h
		mov	cx,600h 	; top line - 6, bottom line - 0
		call	CallBios	; Set cursor size
		mov	ah,12h
		mov	bl,20h
		call	CallBios	; Set print screen for EGA/VGA
l03c6:		ret

SetMode:
		mov	ah,0fh
		call	CallBios	; Getvideo mode
		push	ax		; Saveit
		mov	ax,1130h
		mov	bh,0
		mov	dl,0		; Ptrines:bp, cx=bytes, dl=rows/char
		call	CallBios	; Gettable pointer and info
		pop	ax		; Restore video mode
		mov	cl,0
		or	dl,dl		; rows/char = 0 ?
		jnz	l03e6
		mov	dl,18h		; assume rows := 24 /dec/
		cmp	al,3		; mode=EGA / VGA
		ja	l03e6		; yes
		mov	cl,1		; assumeone byte/char
l03e6:		mov	dh,dl
l03e8:		mov	dl,ah
		dec	dl
		mov	ah,0
		cmp	dh,18h
		jbe	l03f5
		mov	ah,1
l03f5:		mov	LastMode,ax	; savelast mode
		mov	word ptr SaveWindMax,dx
		mov	CheckSnow,cl
		mov	byte ptr DirectVideo,1
		xor	ax,ax
		mov	WindMin,ax
		mov	WindMax,dx
		ret

Int1B:		push	ax
		push	ds
		mov	ax,data
		mov	ds,ax
		cmp	byte ptr CheckBreak,0
		jz	l0422
		mov	byte ptr CtrlBrkFlag,1
l0422:		pop	ds
		pop	ax
		iret

CtrlBrkChk:
		cmp	byte ptr CtrlBrkFlag,0
		jnz	CtrlBrkPending
		ret

CtrlBrkPending:
		mov	byte ptr CtrlBrkFlag,0
FlushBuff:
		mov	ah,1
		int	16h
		jz	Abort
		mov	ah,0
		int	16h
		jmp	short FlushBuff
Abort:
		mov	al,'^'
		call	WriteChar
		mov	al,'C'
		call	WriteChar
		call	WriteLn
		int	23h		; Terminate program

                if      UniCRT

SetTextAttr     proc far		; procedure SetTextAttr(Attr: Byte);
		mov	bx,sp
		mov	al,ss:[bx+4]	; al := Color
		mov	TextAttr, al
		retf	2
SetTextAttr	endp

SetCursor       proc far		; procedure SetTextAttr(n: byte);
		mov	bx,sp
		mov	al,ss:[bx+4]	; al := Cursor size { 0..4 }
                mov     cx, 2000h
                cmp     al, 1
                jnz     Not1crsr
                mov     cx, 0607h
Not1crsr:       cmp     al, 2
                jnz     Not2crsr
                mov     cx, 0407h
Not2crsr:       cmp     al, 3
                jnz     Not3crsr
                mov     cx, 0207h
Not3crsr:       cmp     al, 4
                jnz     Not4crsr
                mov     cx, 0007h
Not4crsr:
                mov     ah, 1
                call    CallBios
		retf	2
SetCursor	endp

                else

LowVideo	proc far		; procedure LowVideo;
		and	TextAttr,not DarkGray	; Reset intensity bit
		retf
LowVideo	endp

HighVideo	proc far		; procedure HighVideo;
		or	TextAttr,DarkGray  ; Set intensity bit
		retf
HighVideo	endp

TextMode	proc far		; procedure TextMode(Mode: Integer);
		mov	bx,sp
		mov	ax,ss:[bx+4]	; ax := Mode
		call	EgaMode
		call	SetMode
		mov	al,SaveMode
		mov	TextAttr,al
		retf	2
TextMode	endp

Window		proc far		; procedure Window(X1,Y1,X2,Y2: Byte);
		mov	bx,sp
		mov	dl,ss:[bx+0ah]	; dl := X1
		mov	dh,ss:[bx+8]	; dh := Y1
		mov	cl,ss:[bx+6]	; cl := X2
		mov	ch,ss:[bx+4]	; ch := Y2
		cmp	dl,cl		; Perform range checks
		ja	WindRet 	; X1 > X2
		cmp	dh,ch
		ja	WindRet 	; Y1 > Y2
		dec	dl
		js	WindRet 	; X1 <= 0
		dec	dh
		js	WindRet 	; Y1 <= 0
		dec	cl
		cmp	cl,byte ptr SaveWindMax   ; X1 > SaveWindMax.X
		ja	WindRet
		dec	ch
		cmp	ch,byte ptr SaveWindMax+1 ; Y1 > SaveWindMax.Y
		ja	WindRet
		mov	WindMin,dx	; Save new windows co-ordinates
		mov	WindMax,cx
		call	SetCursXY	; Position cursor at home
WindRet:
		retf	8
Window		endp

InsLine 	proc far		; procedure InsLine;
		mov	ax,701h 	; Prepare for scroll down
		jmp	short InsDelLine
InsLine 	endp

DelLine 	proc far		; procedure DelLine;
		mov	ax,601h 	; Prepare for scroll up
InsDelLine:
		push	ax
		call	GetCursXY
		pop	ax
		mov	bh,TextAttr
		mov	cl,WindMinX
		mov	ch,dh
		mov	dx,WindMax
		cmp	ch,dh
		jnz	l04f1
		xor	al,al
l04f1:		call	CallBios
		retf
DelLine 	EndP

TextColor	proc far		; procedure TextColor(Color: Byte);
		mov	bx,sp
		mov	al,ss:[bx+4]	; al := Color
		test	al,0f0		; Blinking ?
		jz	NonBlink
		and	al,White
		or	al,Blink
NonBlink:
		and	TextAttr,70h	; Mask Backgrnd attr
		or	TextAttr,al
		retf	2
TextColor	endp

TextBackGround	proc far	       ; procedure TextBackground(Color: Byte);
		mov	bx,sp
		mov	al,ss:[bx+4]	; al := Color
		and	al,LightGray	; Mask Color only
		mov	cl,4
		shl	al,cl		; Move to Backgrnd bits
		and	TextAttr,8fh	; Store in Attr
		or	TextAttr,al
		retf	2
TextBackGround	endp

                endif

ClrScr		proc far		; procedure ClrScr;
		mov	ax,0600h
		mov	bh,TextAttr
		mov	cx,WindMin
		mov	dx,WindMax
		call	CallBios	; Scroll up window
		mov	dx,WindMin
		call	SetCursXY
		retf
ClrScr		endp

ClrEol		proc far		; procedure ClrEol;
		call	GetCursXY
		mov	ax,0600h
		mov	bh,TextAttr
		mov	cx,dx
		mov	dl,WindMaxX
		call	CallBios	; Scroll up window (for one line)
		retf
ClrEol		endp

GotoXY		proc far		; procedure GotoXY(X,Y: Byte);
		mov	bx,sp
		mov	dl,ss:[bx+6]	; dl := X
		mov	dh,ss:[bx+4]	; dl := Y
		dec	dl		; Perform range checks
		add	dl,WindMinX
		jb	GotoRet 	; X < WindMinX
		cmp	dl,WindMaxX
		ja	GotoRet 	; X > WindMaxX
		dec	dh
		add	dh,WindMinY
		jb	GotoRet 	; Y < WindMinX
		cmp	dh,WindMaxY
		ja	GotoRet 	; Y > WinMaxX
		call	SetCursXY
GotoRet:
		retf	4
GotoXY		endp

WhereX		proc far		; function WhereX: Byte;
		call	GetCursXY
		mov	al,dl
		sub	al,WindMinX	; Adjust for current window
		inc	al		; Adjust for 1 .. 80
		retf
WhereX		endp

WhereY		proc far		; function WhereY: Byte;
		call	GetCursXY
		mov	al,dh
		sub	al,WindMinY	; Adjust for current window
		inc	al		; Adjust for 1 .. 25
		retf
WhereY		endp

NormVideo	proc far		; procedure NormVideo;
		mov	al,SaveMode	; Restore original attr
		mov	TextAttr,al
		retf
NormVideo	endp

KeyPressed	proc far		; function KeyPressed: Boolean;
		cmp	CharBuff,0
		jnz	IsPressed
		mov	ah,1
		int	16h
		mov	al,0
		jz	KeyPrsRet
IsPressed:
		mov	al,1
KeyPrsRet:
		retf
KeyPressed	endp

ReadKey 	proc far		; function ReadKey: Char;
		mov	al,CharBuff
		mov	CharBuff,0
		or	al,al
		jnz	ReadKeyRet
		xor	ah,ah
		int	16h
		or	al,al
		jnz	ReadKeyRet
		mov	CharBuff,ah
		or	ah,ah
		jnz	ReadKeyRet
		mov	al,3
ReadKeyRet:
		call	CtrlBrkChk
		retf
ReadKey 	endp

AssignCrt	proc far		; procedure AssignCrt(var F: Text);
		mov	bx,sp
		push	ds
		lds	di,dword ptr ss:[bx+4]	; es:di := F
		mov	word ptr [di+2],fmClosed ; F.Mode := fmClosed
		mov	word ptr [di+4],80	; F.BuffSize := $80
		lea	ax,[di+0080]
		mov	[di+0ch],ax		; F.BuffPtr := @ F.Buff
		mov	[di+0eh],ds
		mov	word ptr [di+10h],offset CrtOpenFunc
		mov	[di+12h],cs		; F.OpenFunc := @ CrtOpenFunc
		mov	byte ptr [di+30h],0	; F.Name := ''
		pop	ds
		retf	4
AssignCrt	endp

CrtOpenFunc	proc far	 ; function CrtOpenFunc(var F: TextRec): Word;
		mov	bx,sp
		push	ds
		lds	di,dword ptr ss:[bx+4]	; es:di := @ F;
		mov	ax,offset CrtInputFunc
		mov	bx,offset CrtDummyFunc
		mov	cx,bx
		cmp	word ptr [di+2],fmInput
		jz	DoOpen
		mov	word ptr [di+2],fmOutput
		mov	ax,offset CrtOutputFunc
		mov	bx,ax
DoOpen:
		mov	[di+14h],ax	; F.InOutFunc := CrtInputFunc  / Input
		mov	[di+16h],cs	; F.InOutFunc := CrtOutputFunc / Output
		mov	[di+18h],bx	; F.FlushFunc := CrtDummyFunc  / Input
		mov	[di+1ah],cs	; F.FlushFunc := CrtOutputFunc / Output
		mov	[di+1ch],cx	; F.CloseFunc := CrtDummyFunc  / Input
		mov	[di+1eh],cs	; F.CloseFunc := CrtDummyFunc  / Output
		xor	ax,ax
		pop	ds
		retf	4
CrtOpenFunc	endp

CrtInputFunc	proc far	; function CrtInputFunc(var F: TextRec): Word;
		push	bp
		mov	bp,sp
		les	di,dword ptr [bp+6]	; es:di := @ F
		mov	dx,es:[di+4]		; dx := F.BuffSize
		dec	dx
		dec	dx			; Decrement - clear CrLf
		mov	si,es:[di+8]		; si := F.BufPos
		les	di,dword ptr es:[di+0ch] ; es:di := F.BufPtr
		xor	bx,bx			; Zero the counter
ReadLoop:
		mov	CharBuff,0
;		push	cs		; Simulate Far Call
;		db	0e8,05ch,0ffh
		call	ReadKey 	; Call is NEAR in original CRT ???
		mov	cx,1		; Loop count for clear entry
		cmp	al,bs		; Back space
		jz	CIFClrEntry
		cmp	al,CtrlS
		jz	CIFClrEntry
		cmp	al,CtrlD
		jz	CIFFilEntry
		dec	cx
		cmp	al,EscCh
		jz	CIFClrEntry
		cmp	al,CtrlA
		jz	CIFClrEntry
		cmp	al,CtrlF
		jz	CIFFilEntry
		cmp	al,CtrlZ
		jz	CIFEofEntry
		cmp	al,cr
		jz	CIFEndEntry
		cmp	al,' '
		jb	ReadLoop
		cmp	bx,dx		; End of buffer ?
		jz	ReadLoop	; Yes, wait for character
		mov	es:[bx+di],al	; Store in the buffer
		inc	bx		; Increment the counter
		call	WriteChar	; Write the char
		cmp	bx,si		; Up to the BufPos
		jbe	ReadLoop	; No
		mov	si,bx
		jmp	short	ReadLoop

CIFClrEntry:					; Clear entry
		or	bx,bx			; Nothing to read
		jz	ReadLoop
		mov	al,bs
		call	WriteChar		; Go back
		mov	al,' '
		call	WriteChar		; Wipe out
		mov	al,bs
		call	WriteChar		; Go back again
		dec	bx
		loop	CIFClrEntry
		jmp	short	ReadLoop

CIFFilEntry:				; Get character from buffer
		cmp	bx,si
		jz	ReadLoop
		mov	al,es:[bx+di]
		cmp	al,' '
		jb	ReadLoop
		call	WriteChar
		inc	bx
		loop	CIFFilEntry
		jmp	short	ReadLoop

CIFEofEntry:				; Ctrl-Z was pressed
		cmp	byte ptr CheckEOF,0
		jz	ReadLoop	; Do not close the input
		mov	es:[bx+di],al
		inc	bx
l0701:		jmp	short CrtInpRet ; Otherwize go out

CIFEndEntry:
		call	WriteLn
		mov	word ptr es:[bx+di],100*lf+cr	; Store LfCr
		inc	bx
		inc	bx
CrtInpRet:
		les	di,dword ptr [bp+6]	; es:di := @ F
		xor	ax,ax
		mov	es:[di+8],ax		; F.BufPos := 0
		mov	es:[di+0ah],bx		; F.BufEnd := bx
		pop	bp
		retf	4
CrtInputFunc	endp

CrtOutputFunc	proc far  ; function CrtOutputFunc(Var F: TextRec) : Word;
		mov	bx,sp
		les	di,dword ptr ss:[bx+4]	; es:di := @ F
		mov	cx,es:[di+8]		; cx := F.BuffPos
		sub	es:[di+8],cx		; F.BufPos := 0
		jcxz	CrtOutRet		; Nothing to write
		les	di,dword ptr es:[di+0ch] ;es.di := F.BuffPtr
		cmp	DirectVideo,0
		jnz	VideoOutput
WriteLoop:					; Loop to write all chars
		mov	al,es:[di]		; Get next char
		call	WriteChar
		inc	di
		loop	WriteLoop
		jmp	short CrtOutRet

VideoOutput:	call	WriteVideoMem

CrtOutRet:	call	CtrlBrkChk
		xor	ax,ax
		retf	4
CrtOutputFunc	endp

CrtDummyFunc	proc	far
		xor	ax,ax
		retf	4
CrtDummyFunc	endp

WriteLn:	mov	al,cr
		call	WriteChar
		mov	al,lf

WriteChar:
		push	bx
		push	cx
		push	dx
		push	es
		push	ax
		call	GetCursXY	; dh = ypos, dl = xpos

		if	UniCRT
		cmp	ESCflag, 0	; check for ESC sequance
		jz	NotESC
		xor	al, al
		mov	ESCflag, al
		pop	ax
		jmp	short RegularChar
NotESC:
		endif

		pop	ax
		cmp	al,bell
		jz	WriteBell
		cmp	al,bs
		jz	WriteBs
		cmp	al,cr
		jz	WriteCr
		cmp	al,lf
		jz	WriteLf

		if	UniCRT
		cmp	al,tab
		jz	WriteTAB
		cmp	al,formfeed
		jz	WriteFF
		cmp	al,clreoln
		jz	WriteCeol
		cmp	al,ESCch
		jnz	NotESCch
		mov	ESCflag, al
		jmp	short WrCharRet
NotESCch:	cmp	al,rubout
		jz	WriteRUB
		endif
RegularChar:
		mov	ah,9
		mov	bl,TextAttr	; Regular char - write it
		xor	bh,bh
		mov	cx,1
		push	dx
		call	CallBios	; Write char / attr at cursor
		pop	dx
		inc	dl		; Increment X pos
		cmp	dl,WindMaxX	; At right edge of window ?
		jbe	WrCharRet	; No, go out

		if	UniCRT		; UniCRT
		dec	dl		; Yes, stay at right edge
		else			; Turbo CRT
		mov	dl,WindMinX	; Otherwise - perform new line
		jmp	short WriteLf
		endif

WriteBell:
		mov	ah,0eh
		call	CallBios	; Write in ASCII mode
		jmp	short WrCharRet

WriteBs:				; Write Back Space
		cmp	dl,WindMinX	; At left edge of window ?
		jz	WrCharRet	; Yes, nothing to do
		dec	dl		; Decrement X pos
		jmp	short WrCharRet ; Go out

WriteCr:	mov	dl,WindMinX	; Move to the left of window
		jmp	short WrCharRet ; Go out

WriteLf:	call	NewLine
		if	UniCRT		; perform CR after LF when UniCRT
		mov	dl,WindMinX	; Move to the left of window
		endif

WrCharRet:
		call	SetCursXY	; Move cursor to new position
		pop	es
		pop	dx
		pop	cx
		pop	bx
		ret

		if	UniCRT
WriteTAB:	jmp	short WriteTAB1
WriteFF:				; clear screen
		mov	ax,0600h
		mov	bh,TextAttr
		mov	cx,WindMin
		mov	dx,WindMax
		call	CallBios	; Scroll up window
		mov	dx,WindMin
		jmp	short WrCharRet
WriteCeol:
		mov	ax,0600h
		mov	bh,TextAttr
		mov	cx,dx
		mov	dl,WindMaxX
		call	CallBios	; Scroll up window (for one line)
		jmp	short WrCharRET

WriteRUB:				; Write rubout
		cmp	dl,WindMinX	; At left edge of window ?
		jz	WrCharRet	; Yes, nothing to do
		dec	dl		; Decrement X pos
		call	WriteSP
		dec	dl
		jmp	short WrCharRet ; Go out

WriteTAB1:	mov	al, dl		;    
		sub	al, WindMinX
		;dec	 al
		and	al, 7
		sub	al, 8
		neg	al
WrSpace:	push	ax
		call	WriteSP
		pop	ax
		cmp	dl,WindMaxX	; At right edge of window ?
		jbe	OK		; No, nothing to do
		mov	dl, WindMaxX
OK:		dec	al
		jnz	WrSpace
		jmp	short WrCharRet

		endif

NewLine:				; Perform new line
		inc	dh		; Increment Y pos
		cmp	dh,WindMaxY	; At bottom edge of window
		jbe	NewLnRet	; No - nothing more to do, go out
		dec	dh		; Restore old value
		push	cx
		push	dx
		mov	ax,601h
		mov	bh,TextAttr
		mov	cx,WindMin
		mov	dx,WindMax
		call	CallBios	; Scroll up window
		pop	dx
		pop	cx
NewLnRet:
		ret

GetCursXY:
		mov	ah,3
		xor	bh,bh
		jmp	CallBios	; Get cursor position

SetCursXY:
		mov	ah,2
		xor	bh,bh
		jmp	CallBios	; Set cursor position

WriteSP:
		call	SetCursXY	; set cursor position
		mov	ax, ' '
		mov	ah,9
		mov	bl,TextAttr	; write ' '
		xor	bh,bh
		mov	cx,1
		push	dx
		call	CallBios	; Write char / attr at cursor
		pop	dx
		inc	dl
		ret

WriteVideoMem	proc near
		push	ds		; Write via Video Memory
		mov	ax,BiosData	; Come with es:di = F.BufPtr
		mov	ds,ax
		mov	dx,CursLocn	; dx := Cursor position
		pop	ds
		mov	bx,dx
		mov	si,di		; es:si := F.BufPtr
l07f2:
		if	UniCRT
		cmp	ESCflag, 0	; check for ESC sequance
		jz	NotESCvideo
		xor	al, al
		mov	ESCflag, al
		mov	al,es:[di]	; Get next character
		jmp	short RegularVideo
NotESCvideo:
		endif

		mov	al,es:[di]	; Get next character
		cmp	al,Bell 	; Bell ?
		jnz	notBELL
		jmp	VideoBell
notBELL:
		cmp	al,bs
		jnz	notBS
		jmp	VideoBS
notBS:
		cmp	al,lf
		jz	VideoLF
		cmp	al,cr
		jz	VideoCR

		if	UniCRT
		cmp	al,tab
		jnz	notTAB
		jmp	WriteTABv
notTAB:
		cmp	al,formfeed
		jnz	notFF
		jmp	WriteFFv
notFF:
		cmp	al,clreoln
		jnz	notCeol
		jmp	WriteCeolv
notCeol:
		cmp	al,rubout
		jnz	notRUB
		jmp	WriteRUBv
notRUB:
		cmp	al,ESCch
		jnz	RegularVideo
		mov	ESCflag, al
		call	MoveScrMem
		jmp	short l0845
		endif
RegularVideo:
		inc	di
		inc	dl
		cmp	dl,WindMaxX
		jbe	l084a
		call	MoveScrMem

		if	UniCRT
		mov	dl,WindMaxX	; stay at right edge
		else
		call	NewLine
		mov	dl,WindMinX
		endif

		jmp	short	l0846

VideoLF:	call	MoveScrMem
		call	NewLine

		if	UniCRT
		jmp	short l084x	; perform CR if UniCRT
		else
		jmp	short l0845
		endif

VideoCR:	call	MoveScrMem

l084x:		mov	dl,WindMinX
l0845:		inc	di
l0846:		mov	si,di
		mov	bx,dx
l084a:		loop	l07f2
		call	MoveScrMem
		push	ds
		mov	ax,BiosData
		mov	ds,ax
		mov	CursLocn,dx
		mov	al,dh
		mul	ScrnWdth
		xor	dh,dh
		add	ax,dx
		mov	cx,ax
		mov	dx,Addr6845
		mov	al,0eh
		out	dx,al
		jmp	short	l086e
l086e:		mov	al,ch
		inc	dx
		out	dx,al
		jmp	short	l0874
l0874:		dec	dx
		mov	al,0fh
		out	dx,al
		jmp	short	l087a
l087a:		mov	al,cl
		inc	dx
		out	dx,al
		pop	ds
		ret

VideoBell:	call	MoveScrMem
		push	cx
		push	dx
		mov	ax,0e07h
		call	CallBios
		pop	dx
		pop	cx
		jmp	short	l0845

VideoBS:	call	MoveScrMem
		cmp	dl,WindMinX
		jz	l0845
		dec	dl
		jmp	short	l0845

		if	UniCRT
WriteTABv:	call	MoveScrMem
		mov	al, ' '
		mov	es:[di], al
		inc	di
		mov	al, dl		;    
		sub	al, WindMinX
		and	al, 7
		sub	al, 8
		neg	al
WrSpaceV:	push	ax
		mov	bx, dx
		call	MoveScrMem
		dec	si
		inc	dl
		pop	ax
		cmp	dl,WindMaxX	; At right edge of window ?
		jbe	OKv		; No, nothing to do
		mov	dl, WindMaxX
OKv:		dec	al
		jnz	WrSpaceV

		dec	di
		mov	al, 9
		mov	es:[di], al
		jmp	l0845

WriteFFv:	call	MoveScrMem
		push	cx
		mov	ax,0600h
		mov	bh,TextAttr
		mov	cx,WindMin
		mov	dx,WindMax
		call	CallBios	; Scroll up window
		mov	dx,WindMin
		pop	cx
		jmp	l0845

WriteCeolv:	call	MoveScrMem
		push	cx
		push	dx
		mov	ax,0600h
		mov	bh,TextAttr
		mov	cx,dx
		mov	dl,WindMaxX
		call	CallBios	; Scroll up window (for one line)
		pop	dx
		pop	cx
		jmp	l0845

WriteRUBv:	call	MoveScrMem
		cmp	dl,WindMinX
		jnz	nottoofar
		jmp	l0845
nottoofar:
		dec	dl
		mov	al, ' '
		mov	es:[di], al
		inc	di
		mov	bx, dx
		call	MoveScrMem
		dec	di
		mov	al, 7fh
		mov	es:[di], al
		jmp	l0845

		endif

WriteVideoMem	endp

MoveScrMem	proc near
		cmp	si,di
		jz	l08e7
		push	cx
		push	dx
		push	di
		push	ds
		push	es
		mov	cx,di
		sub	cx,si
		mov	dl,CheckSnow
		mov	dh,TextAttr
		mov	ax,BiosData
		mov	ds,ax
		mov	al,bh
		mul	ScrnWdth
		xor	bh,bh
		add	ax,bx
		shl	ax,1
		mov	di,ax
		mov	bx,dx
		mov	dx,Addr6845
		add	dx,6
		mov	ax,0b800h
		cmp	CurrMode,7
		jnz	l08bd
		mov	ah,0b0h
l08bd:		push	es
		pop	ds
		mov	es,ax
		cld
		or	bl,bl
		jz	l08dc
l08c6:		lodsb
		mov	bl,al
l08c9:		in	al,dx
		test	al,1
		jnz	l08c9
		cli
l08cf:		in	al,dx
		test	al,1
		jz	l08cf
		mov	ax,bx
		stosw
		sti
		loop	l08c6
		jmp	short	l08e2
l08dc:		mov	ah,bh
l08de:		lodsb
		stosw
		loop	l08de
l08e2:		pop	es
		pop	ds
		pop	di
		pop	dx
		pop	cx
l08e7:		ret
MoveScrMem	endp

CallBios	proc near
		push	si
		push	di
		push	bp
		push	es
		int	10h
		pop	es
		pop	bp
		pop	di
		pop	si
		ret
CallBios	endp

code		ends

		end
